<?php

namespace Ulmus;

use Ulmus\Repository,
    Ulmus\Query,
    Ulmus\Common\EntityResolver,
    Ulmus\Common\EntityField;

use Ulmus\Annotation\Classes\{ Method, Table, Collation, };
use Ulmus\Annotation\Property\{ Field, Filter, Relation, OrderBy, Where, Join, Virtual, On, WithJoin, };
use Ulmus\Annotation\Property\Field\{ Id, ForeignKey, CreatedAt, UpdatedAt, Datetime as DateTime, Date, Time, Bigint, Tinyint, Text, Mediumtext, Longtext, };
use Ulmus\Annotation\Property\Relation\{ Ignore as RelationIgnore };

trait EntityTrait {
    use EventTrait;
    
    /**
     * @Ignore
     */
    protected bool $entityStrictFieldsDeclaration = false;
    
    /**
     * @Ignore
     */
    protected array $entityDatasetUnmatchedFields = [];
    
    /**
     * @Ignore
     */
    public array $entityLoadedDataset = [];

    /**
     * @Ignore
     */
    public function __get(string $name)
    {
        $relation = new Repository\RelationBuilder($this);
        
        if ( false !== $data = $relation->searchRelation($name) ) {
            return $this->$name = $data;
        }
        
        throw new \Exception(sprintf("[%s] - Undefined variable: %s", static::class, $name));
    }
    
    /**
     * @Ignore
     */
    public function __isset(string $name) : bool
    {
        #if ( null !== $relation = static::resolveEntity()->searchFieldAnnotation($name, new Relation() ) ) {
        #    return isset($this->{$relation->key});
        #}
        if ( $this->isLoaded() && static::resolveEntity()->searchFieldAnnotation($name, new Relation() ) ) {
           return true;
        }

        return isset($this->$name);
    }
    
    /**
     * @Ignore
     */
    public function entityFillFromDataset(iterable $dataset) : self
    {
        $loaded = $this->isLoaded();
        
        $entityResolver = $this->resolveEntity();
        
        foreach($dataset as $key => $value) {
            $field = $entityResolver->field(strtolower($key), EntityResolver::KEY_COLUMN_NAME, false) ?? null;
            
            if ( $field === null ) {
                $field = $entityResolver->field(strtolower($key), EntityResolver::KEY_LC_ENTITY_NAME, false);
            }
            
            if ( $field === null ) {
                if ($this->entityStrictFieldsDeclaration ) {
                    throw new \Exception("Field `$key` can not be found within your entity ".static::class);
                }
                else {
                    $this->entityDatasetUnmatchedFields[$key] = $value;
                }
            }
            elseif ( is_null($value) ) {
                $this->{$field['name']} = null;
            }
            elseif ( $field['type'] === 'array' ) {
                if ( is_string($value)) {
                    $this->{$field['name']} = substr($value, 0, 1) === "a" ? unserialize($value) : json_decode($value, true);
                }
                elseif ( is_array($value) ) {
                    $this->{$field['name']} = $value;
                }
            }
            elseif ( EntityField::isScalarType($field['type']) ) {

                if ( $field['type'] === 'string' ) {
                    $annotation = $entityResolver->searchFieldAnnotation($field['name'], new Field() );
                
                    if ( $annotation->length ?? null ) {
                        $value = substr($value, 0, $annotation->length);
                    }
                }
                elseif ( $field['type'] === 'bool' ) {
                    $this->{$field['name']} = (bool) $value;
                }

                $this->{$field['name']} = $value;
            }
            elseif ( ! $field['builtin'] ) {
                $this->{$field['name']} = Ulmus::instanciateObject($field['type'], [ $value ]);
            }
            
            # Keeping original data to diff on UPDATE query
            if ( ! $loaded || $isLoadedDataset ) {
                #if ( $field !== null ) {
                #    $annotation = $entityResolver->searchFieldAnnotation($field['name'], new Field() );
                #    $this->entityLoadedDataset[$annotation ? $annotation->name : $field['name']] = $dataset;       # <--------- THIS TO FIX !!!!!!
                #}
                $this->entityLoadedDataset = array_change_key_case($dataset, \CASE_LOWER);
            }
        }
        
        return $this;
    }

    public function resetVirtualProperties() : self
    {
        foreach($this->resolveEntity()->properties as $prop => $property) {
            if ( ! $property['builtin'] ) {
                foreach($property['tags'] as $tag) {
                    if ( in_array(strtolower($tag['tag']), [ 'relation', 'join' ] ) ) {
                        unset($this->$prop);
                    }
                }
            }
        }

        return $this;
    }
    
    /**
     * @Ignore
     */
    public function fromArray(iterable $dataset) : self
    {
        return $this->entityFillFromDataset($dataset);
    }
    
    /**
     * @Ignore
     */
    public function entityGetDataset(bool $includeRelations = false, bool $returnSource = false) : array
    {
        if ( $returnSource ) {
            return $this->entityLoadedDataset;
        }
        
        $dataset = [];
        
        $entityResolver = $this->resolveEntity();
        
        foreach($entityResolver->fieldList(Common\EntityResolver::KEY_ENTITY_NAME, true) as $key => $field) {
            $annotation = $entityResolver->searchFieldAnnotation($key, new Field() );

            if ( isset($this->$key) ) { 
                $realKey = $annotation->name ?? $key;
                
                switch (true) {
                    case is_object($this->$key):
                        $dataset[$realKey] = Ulmus::convertObject($this->$key);
                    break;
                        
                    case is_array($this->$key):    
                        $dataset[$realKey] = Ulmus::encodeArray($this->$key);
                    break;

                    case is_bool($this->$key):
                        $dataset[$realKey] = (int) $this->$key;
                    break;
                    
                    default:
                       $dataset[$realKey] = $this->$key;
                }
            }
            elseif ( $field['nullable'] ) {
                $dataset[ $annotation->name ?? $key ] = null;
            }
        }
        
        # @TODO Must fix recursive bug !
        if ($includeRelations) {
            foreach($entityResolver->properties as $name => $field){
                $relation = $entityResolver->searchFieldAnnotation($name, new Relation() );
                
                if ( $relation && isset($this->$name) && ($relation->entity ?? $relation->bridge) !== static::class ) {
                    if ( null !== $value = $this->$name ?? null ) {
                        if ( is_iterable($value) ) {
                            $list = [];
                            
                            foreach($value as $entity) {
                                $list[] = $entity->entityGetDataset(false);
                            }
                            
                            $dataset[$name] = $list;
                        }
                        elseif ( is_object($value) ) {
                            $dataset[$name] = $value->entityGetDataset(false);
                        }
                    }
                }
            }
        }  
        
        return $dataset;
    }
    
    /**
     * @Ignore
     */
    public function toArray($includeRelations = false) : array
    {
        return $this->entityGetDataset($includeRelations);
    }
    
    /**
     * @Ignore
     */
    public function toCollection() : EntityCollection
    {
        return static::entityCollection([ $this ]);
    }
    
    /**
     * @Ignore
     */
    public function isLoaded() : bool
    {
        if ( null === $pkField = $this->resolveEntity()->getPrimaryKeyField($this) ) {
            throw new Exception\EntityPrimaryKeyUnknown(sprintf("Entity %s has no field containing attributes 'primary_key'", static::class));
        }
        
        $key = key($pkField);
        
        return isset($this->$key);
    }

    /**
     * @Ignore
     */
    public function __sleep() 
    {
        return array_keys($this->resolveEntity()->fieldList());
    }
    
    /**
     * @Ignore
     */
    public function jsonSerialize()
    {
        return $this->entityGetDataset(); 
    }
    
    /**
     * @Ignore
     */
    public static function resolveEntity() : EntityResolver
    {
        return Ulmus::resolveEntity(static::class);
    }

    /**
     * @Ignore
     */
    public static function repository(string $alias = Repository::DEFAULT_ALIAS) : Repository
    {
        return Ulmus::repository(static::class, $alias);
    }

    /**
     * @Ignore
     */
    public static function entityCollection(...$arguments) : EntityCollection
    {
        $collection = new EntityCollection(...$arguments);
        $collection->entityClass = static::class;

        return $collection;
    }

    /**
     * @Ignore
     */
    public static function field($name, ? string $alias = null) : EntityField
    {
        return new EntityField(static::class, $name, $alias ?: Repository::DEFAULT_ALIAS, Ulmus::resolveEntity(static::class));
    }

    /**
     * @Ignore
     */
    public static function fields(array $fields, ? string $alias = null) : string
    {
        return implode(', ', array_map(function($item) use ($alias){
            return static::field($item, $alias);
        }, $fields));
    }
}
